@@ -1,15 +1,53 @@
-from datetime import datetime
+import os
+import re
 from contextlib import ContextDecorator
+from datetime import datetime
 from enum import Enum
 
 import epics
-import numpy
-
+from epics.ca import current_context, attach_context, ChannelAccessGetFailure
 from gi.repository import GLib, GObject
-from epics.ca import current_context, attach_context
 
 CA_CONTEXT = current_context()
 REUSE = False
+PACKAGE_DIR = os.path.dirname(__file__)
+
+
+def get_version(prefix='v', package=PACKAGE_DIR):
+    from subprocess import CalledProcessError, check_output
+
+    # Return the version if it has been injected into the file by git-archive
+    tag_re = re.compile(rf'\btag: {prefix}([0-9][^,]*)\b')
+    version = tag_re.search('$Format:%D$')
+    name = __name__.split('.')[0]
+
+    if version:
+        return version.group(1)
+
+    package_dir = package
+    if os.path.isdir(os.path.join(package_dir, '.git')):
+        # Get the version using "git describe".
+        version_cmd = 'git describe --tags --abbrev=0'
+        release_cmd = 'git rev-list HEAD ^$(git describe --abbrev=0) | wc -l'
+        try:
+            version = check_output(version_cmd, shell=True).decode().strip()
+            release = check_output(release_cmd, shell=True).decode().strip()
+            return f'{version}.{release}'.strip(prefix)
+        except CalledProcessError:
+            version = '0.0'
+            release = 'dev'
+            return f'{version}.{release}'.strip(prefix)
+    else:
+        try:
+            from importlib import metadata
+        except ImportError:
+            # Running on pre-3.8 Python; use importlib-metadata package
+            import importlib_metadata as metadata
+
+        version = metadata.version(name)
+
+    return version
+
 
 class Alarm(Enum):
     NORMAL, MINOR, MAJOR, INVALID = range(4)
@@ -83,7 +121,6 @@ PV_REPR = (
 )
 
 
-
 class PV(BasePV):
     """A Process Variable
 
@@ -121,23 +158,37 @@ class PV(BasePV):
         self.string = False
 
         # re-use existing instances
-        if REUSE and (name, monitor) in self.__REGISTRY:
-            self.raw = self.__REGISTRY[(name, monitor)]
+        if REUSE and name in self.__REGISTRY:
+            self.raw = self.__REGISTRY[name]
+            self.raw.add_callback(callback=self.on_change, with_ctrlvars=True)
+            self.raw.connection_callbacks.append(self.on_connect)
+            GLib.timeout_add(100, self.update_state)    # make sure we update the state if it is already connected
         else:
-            self.raw = epics.PV(name, callback=self.on_change, connection_callback=self.on_connect, auto_monitor=monitor)
-            self.__REGISTRY[(name, monitor)] = self.raw
+            self.raw = epics.PV(name, auto_monitor=True, callback=self.on_change, connection_callback=self.on_connect)
+            self.__REGISTRY[name] = self.raw
+
+    def update_state(self,):
+        """
+        Update the state of the PV and emit signals.
+        This method is called when the PV is connected or its value changes.
+        """
+
+        if self.raw.status == 0:
+            value = self.raw.char_value if self.string else self.raw.value
+            alarm = Alarm(self.raw.severity)
+            self.set_state(active=True, changed=value, time=datetime.fromtimestamp(self.raw.timestamp), alarm=alarm)
 
     def on_connect(self, **kwargs):
         self.set_state(active=kwargs['conn'])
 
     def on_change(self, **kwargs):
-        self.string = kwargs['type'] in ['time_string', 'time_char']
+        self.string = kwargs['type'] in ['time_string'] or (kwargs['type'] in ['time_char'] and kwargs['count'] > 1)
         value = kwargs['char_value'] if self.string else kwargs['value']
         alarm = Alarm(kwargs.get('severity', 0))
         self.set_state(changed=value, time=datetime.fromtimestamp(kwargs['timestamp']), alarm=alarm)
 
     def get(self, *args, **kwargs):
-        kwargs['as_string'] = self.string
+        kwargs['as_string'] |= self.string
         return self.raw.get(*args, **kwargs)
 
     def put(self, *args, **kwargs):
@@ -152,6 +203,8 @@ class PV(BasePV):
             return getattr(self.raw, item)
         except AttributeError:
             raise AttributeError('%r object has no attribute %r' % (self.__class__.__name__, item))
+        except ChannelAccessGetFailure:
+            return None
 
     def __repr__(self):
         return PV_REPR.format(
